﻿using System;
using System.Collections.Generic;

namespace jello
{
    class Commands
    {
        private List<Command> CommandsList = new List<Command> { };

        private int FindLabel(string label, ref InternalState state)
        {
            int counter = 0;
            while (counter < state.CodeLines.Count)
            {
                if (state.CodeLines[counter].TrimStart(new char[] { ' ', '\n', '\t', '\r' }) == label)
                    return counter;
                counter += 1;
            }
            return 0;
        }

        private int FindNextEnd(ref InternalState state)
        {
            int counter = state.Pointer;
            while (counter < state.CodeLines.Count)
            {
                if (state.CodeLines[counter].TrimStart(new char[] { ' ', '\n', '\t', '\r' }) == "END")
                    return counter;
                counter += 1;
            }
            return 0;
        }

        public Commands()
        {
            //Command: ADD
            CommandCode code = delegate(string[] vals, ref InternalState state)
            {
                state.Register = (short)(state.Register + ValueInterpreter.Interpret(vals[0], ref state));
                return 1;
            };
            CommandsList.Add(new Command("ADD", 1, code));

            //Command: SUB
            code = delegate(string[] vals, ref InternalState state)
            {
                state.Register = (short)(state.Register - ValueInterpreter.Interpret(vals[0], ref state));
                return 1;
            };
            CommandsList.Add(new Command("SUB", 1, code));

            //Command: DIV
            code = delegate(string[] vals, ref InternalState state)
            {
                state.Register = (short)(state.Register / ValueInterpreter.Interpret(vals[0], ref state));
                return 1;
            };
            CommandsList.Add(new Command("DIV", 1, code));

            //Command: MUL
            code = delegate(string[] vals, ref InternalState state)
            {
                state.Register = (short)(state.Register * ValueInterpreter.Interpret(vals[0], ref state));
                return 1;
            };
            CommandsList.Add(new Command("MUL", 1, code));

            //Command: SETREG
            code = delegate(string[] vals, ref InternalState state)
            {
                state.Register = (short)(ValueInterpreter.Interpret(vals[0], ref state));
                return 1;
            };
            CommandsList.Add(new Command("SETREG", 1, code));

            //Command: SETMEM
            code = delegate(string[] vals, ref InternalState state)
            {
                state.SetMemory((short)(ValueInterpreter.Interpret(vals[0], ref state)), (short)(ValueInterpreter.Interpret(vals[1], ref state)));
                return 1;
            };
            CommandsList.Add(new Command("SETMEM", 2, code));

            //Command: SETOUT
            code = delegate(string[] vals, ref InternalState state)
            {
                state.Output = (short)(ValueInterpreter.Interpret(vals[0], ref state));
                Console.WriteLine("  -> OUT: " + state.Output.ToString());
                return 1;
            };
            CommandsList.Add(new Command("SETOUT", 1, code));

            //Command: GETINP
            code = delegate(string[] vals, ref InternalState state)
            {
                short val = 0;
                Console.Write("  -> INPUT: ");
                if(short.TryParse(Console.ReadLine(), out val))
                    state.Input = val;
                else
                    state.Input = 0;
                return 1;
            };
            CommandsList.Add(new Command("GETINP", 0, code));

            //Command: END
            code = delegate(string[] vals, ref InternalState state)
            {
                return 1;
            };
            CommandsList.Add(new Command("END", 0, code));

            //Command: GOTO
            code = delegate(string[] vals, ref InternalState state)
            {
                if (vals[0][0] == ':')
                {
                    return FindLabel(vals[0], ref state) - state.Pointer;
                }
                else if (vals[0][0] == '+' || vals[0][0] == '-')
                {
                    return ValueInterpreter.Interpret(vals[0], ref state);
                }
                else
                {
                    return ValueInterpreter.Interpret(vals[0], ref state) - state.Pointer - 1;
                }
            };
            CommandsList.Add(new Command("GOTO", 1, code));

            //Command: IF with 1 input
            code = delegate(string[] vals, ref InternalState state)
            {
                if (vals[0] == "ZERO")
                {
                    if (state.Register == 0)
                        return 1;
                    else
                        return FindNextEnd(ref state) - state.Pointer;
                }
                return 1;
            };
            CommandsList.Add(new Command("IF", 1, code));

            //Command: IF with 2 inputs
            code = delegate(string[] vals, ref InternalState state)
            {
                if (vals[0] == "EQ")
                {
                    if (state.Register == ValueInterpreter.Interpret(vals[1], ref state))
                        return 1;
                    else
                        return FindNextEnd(ref state) - state.Pointer;
                }
                else if (vals[0] == "GT")
                {
                    if (state.Register > ValueInterpreter.Interpret(vals[1], ref state))
                        return 1;
                    else
                        return FindNextEnd(ref state) - state.Pointer;
                }
                else if (vals[0] == "LT")
                {
                    if (state.Register < ValueInterpreter.Interpret(vals[1], ref state))
                        return 1;
                    else
                        return FindNextEnd(ref state) - state.Pointer;
                }
                else if (vals[0] == "UN")
                {
                    if (state.Register != ValueInterpreter.Interpret(vals[1], ref state))
                        return 1;
                    else
                        return FindNextEnd(ref state) - state.Pointer;
                }
                return 1;
            };
            CommandsList.Add(new Command("IF", 2, code));
        }

        public int RunCommand(string command, string[] values, ref InternalState State)
        {
            foreach (Command item in CommandsList)
            {
                if (item.NumberOfValues == values.Length && item.Name == command)
                    return item.Run(values, ref State);
            }
            return 1;
        }
    }
}
